"""Utilities for buildthings unit tests.

Version Added:
    1.1
"""

from __future__ import annotations

import os
import shutil
import sys
import tempfile
from contextlib import contextmanager
from importlib.metadata import Distribution
from typing import TYPE_CHECKING
from zipfile import ZipFile

from pyproject_hooks import BuildBackendHookCaller

import buildthings

if TYPE_CHECKING:
    from collections.abc import Iterator


_buildthings_path = os.path.dirname(buildthings.__file__)


@contextmanager
def setup_pyproject_toml(
    content: str,
) -> Iterator[None]:
    """Set up a temp working directory with a pyproject.toml file.

    This is a context manager that unit tests can use to establish a working
    package directory for testing. It will write the provided content to a
    :file:`pyproject.toml` in the directory and then run the test within it.
    The directory and all contents will be cleaned up after the test.

    Version Added:
        1.1

    Args:
        content (str):
            The content to write to the :file:`pyproject.toml` file.

    Context:
        The test will run within the populated temp directory.
    """
    prev_cwd = os.getcwd()
    dirpath = tempfile.mkdtemp()

    os.chdir(dirpath)

    with open('pyproject.toml', 'w', encoding='utf-8') as fp:
        fp.write("""
            [build-system]
            requires = ['buildthings']
            build-backend = 'buildthings.backend'

            [project]
            name = 'buildthings-test'
            version = '1.0'
        """)
        fp.write(content)

    try:
        yield
    finally:
        os.chdir(prev_cwd)
        shutil.rmtree(dirpath)


@contextmanager
def setup_backend(
    pyproject_toml: str,
) -> Iterator[BuildBackendHookCaller]:
    """Set up a backend environment and hook caller for a test.

    This will set up a tree with a :file:`pyproject.toml` and then configure
    a backend hook caller so that a hook can be invoked in a subprocess.

    Version Added:
        1.1

    Args:
        pyproject_toml (str):
            The content to write to the :file:`pyproject.toml` file.

    Context:
        pyproject_hooks.BuildBackendHookCaller:
        The test will run within the populated temp directory with a backend
        hook caller instance.
    """
    with setup_pyproject_toml(pyproject_toml):
        os.symlink(_buildthings_path, 'buildthings')

        yield BuildBackendHookCaller(
            source_dir=os.getcwd(),
            build_backend='buildthings.backend',
            backend_path='.',
            runner=None,
            python_executable=sys.executable,
        )


def get_dist_from_wheel(
    path: str,
) -> Distribution:
    """Return a Distribution from a wheel file.

    This assumes a specific name and version for the package.

    Version Added:
        1.1

    Args:
        path (str):
            The path to the wheel.

    Returns:
        importlib.metadata.Distribution:
        The resulting distribution.
    """
    with ZipFile(path) as zf:
        zf.extractall('.wheel')

    dist = Distribution.at('.wheel/buildthings_test-1.0.dist-info')
    assert dist.name

    return dist



"""pyproject.toml configuration deserialization.

Version Added:
    1.0
"""

from __future__ import annotations

import os
import sys
from importlib import import_module
from typing import Literal, TYPE_CHECKING, TypedDict

from buildthings.requirements import parse_requirements_file

try:
    # Python >= 3.11
    from tomllib import load as load_toml
    _read_mode = 'rb'
except ImportError:
    # Python <= 3.10
    from toml import load as load_toml
    _read_mode = 'r'

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any, Callable, NotRequired

    from typing_extensions import Self, TypeAlias

    _PyProjectConfigDict: TypeAlias = Mapping[str, Any]


#: A supported build type.
#:
#: Version Added:
#:     1.0
BuildType: TypeAlias = Literal['editable', 'sdist', 'wheel']


#: All supported build types.
#:
#: Version Added:
#:     1.0
ALL_BUILD_TYPES: Sequence[BuildType] = ['editable', 'sdist', 'wheel']


class BuildStep(TypedDict):
    """A custom build step for the package.

    Version Added:
        1.0
    """

    #: The command to run.
    command: str

    #: The optional label to display when running the command.
    label: NotRequired[str]


class BuildPackageDepsConfig(TypedDict):
    """Configuration for a build's package dependencies.

    Version Added:
        1.1
    """

    #: The dependencies for the build type.
    #:
    #: This contains the build-specific dependencies, if provided, or the
    #: common dependencies if not.
    #:
    #: It does not factor in any exclusions from :py:attr:`exclude_deps`.
    dependencies: Sequence[str]

    #: The dependencies to exclude.
    #:
    #: These must be removed from :py:attr:`dependencies` when building a
    #: final dependencies list.
    exclude_deps: Sequence[str]


class IsolationConfig(TypedDict):
    """Configuration for an isolated build.

    Version Added:
        1.0
    """

    #: The list of dependencies to exclude from the isolated environment.
    exclude_deps: Sequence[str]

    #: Whether to include dev dependencies in the isolated environment.
    #:
    #: If set, all of the package's dependencies will be included as
    #: development dependencies when setting up the isolated environment.
    include_dev_deps: bool

    #: Whether to include package dependencies in the isolated environment.
    #:
    #: If set, all of the package's dependencies will be included as
    #: install-time dependencies when setting up the isolated environment.
    include_install_deps: bool

    #: The path to the local development Python packages.
    #:
    #: These are used for installing local packages in an isolated
    #: environment.
    #:
    #: If unset, local packages won't be used for the build type.
    local_packages_path: str | None


class NPMConfig(TypedDict):
    """Configuration for NPM.

    Version Added:
        1.0
    """

    #: Whether NPM packages should be installed for the build type.
    install: bool

    #: The list of Python modules required for NPM installs.
    #:
    #: Each of these will be set up as an NPM workspace, allowing any
    #: :file:`package.json` files in these modules to be used as
    #: dependencies for generating a :file:`node_modules` and running
    #: custom build steps.
    python_modules: Sequence[str]


class PyProjectConfig:
    """A loaded pyproject.toml configuration file.

    This represents a :file:`pyproject.toml` file, with any buildthings
    settings loaded and deserialized, ready to build the package.

    Version Added:
        1.0
    """

    ######################
    # Instance variables #
    ######################

    #: The full deserialized pyproject.toml data.
    config: _PyProjectConfigDict

    #: The list of editable-time development dependencies.
    dev_dependencies: Sequence[str]

    #: The set of all dynamic keys.
    dynamic: set[str]

    #: The editable mode used for the package.
    editable_mode: Literal['compat', 'lenient', 'strict']

    #: A list of extra build steps for each build type.
    extra_build_steps: Mapping[BuildType, Sequence[BuildStep]]

    #: The list of all files to collect for the package.
    files_to_collect: list[str]

    #: Build-specific configuration for isolated environments.
    isolation_build_configs: Mapping[BuildType, IsolationConfig]

    #: Build-specific configuration for NPM.
    npm_configs: Mapping[BuildType, NPMConfig]

    #: The list of install-time dependencies.
    #:
    #: Version Added:
    #:     1.1
    package_dependencies: Mapping[BuildType, BuildPackageDepsConfig]

    @classmethod
    def from_file(
        cls,
        path: (str | None) = None,
    ) -> Self:
        """Return a loaded project configuration from a file.

        Args:
            path (str, optional):
                The path to load.

                If not provided, :file:`pyproject.toml` will be loaded.

        Returns:
            PyProjectConfig:
            The loaded configuration.
        """
        config = cls()
        config.load(path)

        return config

    def __init__(self) -> None:
        """Initialize the project configuration."""
        self.config = {}
        self.files_to_collect = []
        self.dynamic = set()
        self.dev_dependencies = []
        self.editable_mode = 'compat'
        self.extra_build_steps = {}
        self.isolation_build_configs = {}
        self.npm_configs = {}
        self.package_dependencies = {}

    def load(
        self,
        path: (str | None) = None,
    ) -> None:
        """Load a pyproject.toml file.

        This will attempt to load the :file:`pyproject.toml` file and parse
        it for any configuration needed for building the package.

        All configuration keys found in the file will be made available
        to any callers.

        Args:
            path (str, optional):
                The path to load.

                If not provided, :file:`pyproject.toml` will be loaded.
        """
        if not path:
            path = os.path.join(os.getcwd(), 'pyproject.toml')

        sys.path.insert(0, os.path.dirname(path))

        with open(path, _read_mode) as fp:
            self.config = load_toml(fp)

        # Load the common state.
        self._load_dynamic()
        self._load_dev_dependencies()

        common_deps = self._load_dependencies()

        # Isolation configuration.
        isolation_common_config = self._load_isolation_options(
            build_type=None,
            default=IsolationConfig(
                exclude_deps=[],
                include_dev_deps=False,
                include_install_deps=False,
                local_packages_path=None,
            ),
        )

        isolation_defaults = {
            'editable': IsolationConfig(**{
                **isolation_common_config,
                'local_packages_path': '.local-packages',
            }),
        }

        self.isolation_build_configs = {
            build_type: self._load_isolation_options(
                build_type=build_type,
                default=isolation_defaults.get(build_type,
                                               isolation_common_config),
            )
            for build_type in ALL_BUILD_TYPES
        }

        # Per-build package dependency configuration.
        self.package_dependencies = {
            build_type: BuildPackageDepsConfig(
                dependencies=self.load_config_value(
                    full_key=f'tool.buildthings.{build_type}.dependencies',
                    default=common_deps,
                    validate_type=list,
                ),
                exclude_deps=self.load_config_value(
                    full_key=f'tool.buildthings.{build_type}.exclude-deps',
                    default=[],
                    validate_type=list,
                ),
            )
            for build_type in ALL_BUILD_TYPES
        }

        # Build steps configuration.
        self.extra_build_steps = {
            build_type: self.load_config_value(
                full_key=f'tool.buildthings.{build_type}.extra-build-steps',
                default=[],
            )
            for build_type in ALL_BUILD_TYPES
        }

        # Editable configuration.
        editable_mode = self.load_config_value(
            full_key='tool.buildthings.editable.install-mode',
            default=self.editable_mode,
        )

        if editable_mode not in {'compat', 'lenient', 'strict'}:
            raise ValueError(
                'tool.buildthings.editable.install-mode must be one of: '
                'compat, lenient, strict'
            )

        self.editable_mode = editable_mode

        # NPM configuration.
        self.npm_configs = {
            build_type: self._load_npm_config(build_type=build_type)
            for build_type in ALL_BUILD_TYPES
        }

    def load_config_value(
        self,
        *,
        full_key: str,
        allow_dynamic: bool = True,
        default: Any = None,
        file_loader: (Callable[[str], Any] | None) = None,
        validate_type: (type | None) = None,
    ) -> Any:
        """Load a value from a configuration key.

        This can load a key specified as a dotted path within the
        deserialized configuration data.

        If the key is marked as a dynamic configuration key, this will
        take care of loading the corresponding value either from the
        specified attribute or file.

        Version Changed:
            1.1:
            Added the ``validate_type`` argument.

        Args:
            full_key (str):
                The full key to load.

            allow_dynamic (bool, optional):
                Whether this key's value is allowed to be specified
                dynamically.

            default (object, optional):
                The default value to return if a key was not found.

            file_loader (callable, optional):
                The file loader used to load dynamic file specifiers, if any.

            validate_type (type, optional):
                A type that the result must match.

                If provided, and the type is not a match, a
                :py:exc:`ValueError` will be raised with details.

                Version Added:
                    1.1

        Returns:
            object:
            The loaded value.

        Raises:
            ValueError:
                A configuration setting did not meet expectations.

                Details will be in the error message.
        """
        key_parts = full_key.split('.')
        key = key_parts[-1]

        try:
            config_value = self._load_key_dotted(self.config, key_parts)
        except KeyError:
            return default

        is_dynamic = key in self.dynamic

        if not is_dynamic:
            if (validate_type is not None and
                not isinstance(config_value, validate_type)):
                raise ValueError(
                    f'Key "{full_key}" must resolve to a '
                    f'{validate_type.__name__} (not '
                    f'{type(config_value).__name__}).'
                )

            return config_value

        if not allow_dynamic:
            raise ValueError(
                f'Key "{key}" cannot be specified in project.dynamic='
            )

        if not isinstance(config_value, dict):
            raise ValueError(
                f'Dynamic key "{full_key}" must be a dictionary with "attr" '
                f'or "file" keys in pyproject.toml.'
            )

        has_attr = 'attr' in config_value
        has_file = 'file' in config_value

        if not has_attr and not has_file:
            # This is missing an "attr" or "file" key.
            raise ValueError(
                f'Missing key "attr" or "file" for dynamic key '
                f'"{full_key}" in pyproject.toml.'
            )
        elif has_attr and has_file:
            # This includes both an "attr" and "file" key.
            raise ValueError(
                f'Only one of "attr" or "file" may be specified for '
                f'dynamic key "{full_key}" in pyproject.toml.'
            )
        elif has_attr:
            # There's an "attr" key. Load the specified object and attribute
            # represented by the Python path.
            attr_value = config_value['attr']
            mod_path, attr_name = attr_value.rsplit('.', 1)

            try:
                module = import_module(mod_path)
            except Exception as e:
                raise ValueError(
                    f'Unable to load dynamic module attribute "{attr_value}": '
                    f'{e}'
                )

            result = getattr(module, attr_name)

            if result is not None and callable(result):
                result = result()

            if (validate_type is not None and
                not isinstance(result, validate_type)):
                raise ValueError(
                    f'Key "{full_key}" must resolve to a '
                    f'{validate_type.__name__} (not '
                    f'{type(result).__name__}).'
                )

            return result
        elif has_file:
            # Theres a "file" key. Load the file and run it through the
            # specified parser.
            assert file_loader is not None

            file_value = os.path.abspath(config_value['file'])

            try:
                return self._load_file(
                    path=file_value,
                    file_loader=file_loader,
                )
            except IOError as e:
                raise IOError(
                    f'Unable to read "{file_value}" (specified as '
                    f'key "{full_key}" in pyproject.toml): {e}:'
                )

        raise AssertionError('Not reached')

    def _load_file(
        self,
        path: str,
        *,
        file_loader: Callable[[str], Any],
    ) -> Any:
        """Load contents from a file.

        This will attempt to load a file and run it through the provided
        file loader. If successful, the file will be marked as a file to
        collect when packaging.

        It's up to the caller to check for any exceptions raised by the
        file loader.

        Args:
            path (str):
                The file path to load.

            file_loader (callable):
                The function used to load and return contents from the file.

        Raises:
            Exception:
                Any exception raised by the file loader.
        """
        result = file_loader(path)
        self.files_to_collect.append(path)

        return result

    def _load_dynamic(self) -> None:
        """Load the set of all dynamic configuration keys.

        This will look in ``tool.buildthings.dynamic`` for configuration keys
        that should be looked up through a Python import.
        """
        dynamic: set[str] = set()

        try:
            dynamic.update(self._load_key_dotted(
                self.config,
                ('project', 'dynamic'),
            ))
        except KeyError:
            pass

        try:
            dynamic.update(self._load_key_dotted(
                self.config,
                ('tool', 'buildthings', 'dynamic'),
            ))
        except KeyError:
            pass

        self.dynamic = dynamic

    def _load_dependencies(self) -> Sequence[str]:
        """Load a list of dependencies.

        This is loaded from ``tool.buildthings.dependencies``. If not found,
        this will look for a :file:`requirements.txt` and try to load from
        that.

        Version Changed:
            1.1:
            This now returns the dependencies instead of setting them.

        Returns:
            list of str:
            The list of dependencies.

        Raises:
            ValueError:
                A dependency configuration setting did not meet expectations.

                Details will be in the error message.
        """
        result = self.load_config_value(
            full_key='tool.buildthings.dependencies',
            file_loader=parse_requirements_file,
        )

        if result is None:
            try:
                result = self._load_file(
                    'requirements.txt',
                    file_loader=parse_requirements_file,
                )
            except IOError:
                # We were optimistically looking for this, so just ignore
                # the error.
                result = []
        elif not isinstance(result, list):
            raise ValueError(
                f'Key "tool.buildthings.dependencies" must resolve to a '
                f'list of dependencies (not {result!r}). Did you forget to '
                f'list this in "tool.buildthings.dynamic"?'
            )

        return result or []

    def _load_dev_dependencies(self) -> None:
        """Load a list of development dependencies.

        These are meant for use at editable time.

        This is loaded from ``tool.buildthings.dev-dependencies``. If not
        found, this will look for a :file:`dev-requirements.txt` and try
        to load from that.

        Raises:
            ValueError:
                A dependency configuration setting did not meet expectations.

                Details will be in the error message.
        """
        result = self.load_config_value(
            full_key='tool.buildthings.dev-dependencies',
            file_loader=parse_requirements_file,
        )

        if result is None:
            try:
                result = self._load_file(
                    'dev-requirements.txt',
                    file_loader=parse_requirements_file,
                )
            except IOError:
                # We were optimistically looking for this, so just ignore
                # the error.
                result = []
        elif not isinstance(result, list):
            raise ValueError(
                f'Key "tool.buildthings.dev-dependencies" must resolve to a '
                f'list of dependencies (not {result!r}). Did you forget to '
                f'list this in "tool.buildthings.dynamic"?'
            )

        self.dev_dependencies = result

    def _load_isolation_options(
        self,
        *,
        build_type: str | None,
        default: IsolationConfig,
    ) -> IsolationConfig:
        """Load configuration for an isolated environment.

        Args:
            build_type (str):
                The build type, or ``None`` to reference common options.

            default (IsolationConfig):
                The default values for each option.

        Returns:
            IsolationConfig:
            The loaded configuration.
        """
        prefix = 'tool.buildthings.isolation'

        if build_type:
            prefix = f'{prefix}.{build_type}'

        return IsolationConfig(
            exclude_deps=self.load_config_value(
                full_key=f'{prefix}.exclude-deps',
                default=default['exclude_deps'],
            ),
            include_dev_deps=self.load_config_value(
                full_key=f'{prefix}.include-dev-deps',
                default=default['include_dev_deps'],
            ),
            include_install_deps=self.load_config_value(
                full_key=f'{prefix}.include-install-deps',
                default=default['include_install_deps'],
            ),
            local_packages_path=self.load_config_value(
                full_key=f'{prefix}.local-packages-path',
                default=default['local_packages_path'],
            ),
        )

    def _load_npm_config(
        self,
        *,
        build_type: BuildType,
    ) -> NPMConfig:
        """Load configuration for NPM.

        Args:
            build_type (str):
                The build type.

        Returns:
            NPMConfig:
            The loaded configuration.
        """
        prefix = f'tool.buildthings.{build_type}.npm'

        return NPMConfig(
            install=self.load_config_value(
                full_key=f'{prefix}.install',
                default=False,
            ),
            python_modules=self.load_config_value(
                full_key=f'{prefix}.python-modules',
                default=[],
            ),
        )

    def _load_key_dotted(
        self,
        d: Mapping[str, Any],
        keys: Sequence[str],
    ) -> Any:
        """Return a value from a mapping given a dotted path.

        Args:
            d (dict):
                The dictionary to load from.

            keys (list of str):
                The list of key path components.

        Returns:
            object:
            The resulting value.
        """
        for key in keys:
            d = d[key]

        return d



"""Buildthings Python build backend.

Version Added:
    1.0
"""

from buildthings._version import (VERSION,
                                  __version__,
                                  __version_info__,
                                  get_package_version,
                                  get_version_string,
                                  is_release)


__all__ = [
    'VERSION',
    '__version__',
    '__version_info__',
    'get_package_version',
    'get_version_string',
    'is_release',
]



"""Tests for buildthings.config.Config.

Version Added:
    1.1
"""

from __future__ import annotations

import pytest

from buildthings.config import PyProjectConfig
from buildthings.tests.utils import setup_pyproject_toml


def test_load_dependencies() -> None:
    """Test PyProjectConfig.load with dependencies.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'ReviewBoard>=8.0',
            'buildthings',
            'kgb~=7.0',
        ]
    """

    with setup_pyproject_toml(pyproject_toml):
        config = PyProjectConfig.from_file('pyproject.toml')

    assert config.package_dependencies == {
        'editable': {
            'dependencies': [
                'ReviewBoard>=8.0',
                'buildthings',
                'kgb~=7.0',
            ],
            'exclude_deps': [],
        },
        'sdist': {
            'dependencies': [
                'ReviewBoard>=8.0',
                'buildthings',
                'kgb~=7.0',
            ],
            'exclude_deps': [],
        },
        'wheel': {
            'dependencies': [
                'ReviewBoard>=8.0',
                'buildthings',
                'kgb~=7.0',
            ],
            'exclude_deps': [],
        },
    }


def test_load_dependencies_with_build_type_overrides() -> None:
    """Test PyProjectConfig.load with dependencies/exclude-deps with
    build type overrides.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'ReviewBoard>=8.0',
            'buildthings',
            'kgb~=7.0',
        ]

        [tool.buildthings.editable]
        dependencies = ['my-editable-dep']
        exclude-deps = ['exclude-editable-dep']

        [tool.buildthings.sdist]
        dependencies = ['my-sdist-dep']
        exclude-deps = ['exclude-sdist-dep']

        [tool.buildthings.wheel]
        dependencies = ['my-wheel-dep']
        exclude-deps = ['exclude-wheel-dep']
    """

    with setup_pyproject_toml(pyproject_toml):
        config = PyProjectConfig.from_file('pyproject.toml')

    assert config.package_dependencies == {
        'editable': {
            'dependencies': ['my-editable-dep'],
            'exclude_deps': ['exclude-editable-dep'],
        },
        'sdist': {
            'dependencies': ['my-sdist-dep'],
            'exclude_deps': ['exclude-sdist-dep'],
        },
        'wheel': {
            'dependencies': ['my-wheel-dep'],
            'exclude_deps': ['exclude-wheel-dep'],
        },
    }


def test_load_dependencies_with_non_list() -> None:
    """Test PyProjectConfig.load with dependencies with non-list.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = 'XXX'
    """

    with setup_pyproject_toml(pyproject_toml):
        with pytest.raises(ValueError) as excinfo:
            PyProjectConfig.from_file('pyproject.toml')

    assert str(excinfo.value) == (
        'Key "tool.buildthings.dependencies" must resolve to a list of '
        'dependencies (not \'XXX\'). Did you forget to list this in '
        '"tool.buildthings.dynamic"?'
    )


def test_load_dependencies_override_with_non_list() -> None:
    """Test PyProjectConfig.load with dependencies override with non-list.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings.editable]
        dependencies = 'XXX'
    """

    with setup_pyproject_toml(pyproject_toml):
        with pytest.raises(ValueError) as excinfo:
            config = PyProjectConfig.from_file('pyproject.toml')

            print(config.package_dependencies)

    assert str(excinfo.value) == (
        'Key "tool.buildthings.editable.dependencies" must resolve to a '
        'list (not str).'
    )



"""Parsing support for requirements.txt-like files.

Version Added:
    1.0
"""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from packaging.requirements import InvalidRequirement, Requirement
from packaging.utils import canonicalize_name

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


#: A regex for parsing comments in a requirements.txt.
#:
#: Version Added:
#:     1.0
COMMENT_RE = re.compile(r'#.*$')


def get_package_name_from_specifier(
    dep: str,
) -> str:
    """Return the name of a package from a dependency specifier.

    If the specifier includes extras (e.g., ``ReviewBoard[p4]``), then
    that will be included in the result for the package name.

    If the package can't be parsed, the name will be returned as-is and the
    failure will be logged.

    Version Added:
        1.1

    Args:
        dep (str):
            The dependency specifier.

    Returns:
        str:
        The package name.
    """
    try:
        requirement = Requirement(dep)
    except InvalidRequirement:
        logger.warning('Invalid dependency name %r', dep)

        return dep

    # We ensure a str because canonicalize_name returns a more specific (but
    # otherwise compatible) type for a str.
    name: str = canonicalize_name(requirement.name)

    if extras := requirement.extras:
        extras_str = ','.join(sorted(
            canonicalize_name(extra)
            for extra in extras
        ))
        name = f'{name}[{extras_str}]'

    return name


def filter_dependencies(
    deps: Sequence[str],
    *,
    exclude: set[str],
) -> list[str]:
    """Return a set of dependencies with specified packages excluded.

    This will process the provided list of dependencies and remove any that
    are found in ``exclude``.

    Package names are normalized for comparison, using the canonical package
    name rather than the form specified in either list.

    The exclude list is meant to be treated as a set of package names rather
    than specific versions. Any version specifiers will be stripped during
    comparison.

    Package names in either the list of dependencies or the exclude list may
    support ``[extras]``, but these must be an exact match. That is,
    ``reviewboard[p4]`` will match ``ReviewBoard[p4]``, but
    ``reviewboard[p4,ldap]`` will *not* match either ``ReviewBoard[p4]`` or
    ``ReviewBoard[ldap]``.

    Version Added:
        1.1

    Args:
        deps (list of str):
            The list of dependencies to filter.

        exclude (set of str):
            The dependencies to exclude.

    Returns:
        list of str:
        The resulting list of dependencies.
    """
    # Normalize all excluded dependency names.
    exclude_deps = {
        get_package_name_from_specifier(dep)
        for dep in exclude
    }

    return [
        dep
        for dep in deps
        if get_package_name_from_specifier(dep) not in exclude_deps
    ]


def parse_requirements_file(
    path: str,
) -> Sequence[str]:
    """Return valid dependency strings from a requirements.txt file.

    Only valid dependency specifiers will be returned. Items starting
    with ``-e``, ``-r``, `--index-url``, etc. will be skipped.

    Version Added:
        1.0

    Args:
        path (str):
            The path to the requirements file.

    Returns:
        list of str:
        The list of sanitized dependency specifiers.
    """
    with open(path, 'r', encoding='utf-8') as fp:
        return [
            parsed
            for line in fp
            if (parsed := _parse_requirement_line(line, path))
        ]


def _parse_requirement_line(
    line: str,
    path: str,
) -> str:
    """Parse a line from a requirements.txt file.

    This will remove any comments, validate the line as a valid requirement,
    and then return it if it passes validation.

    Version Added:
        1.0

    Args:
        line (str):
            The line to parse.

        path (str):
            The path to the file being processed.

    Returns:
        str:
        The sanitized requirement line, or an empty string if it fails to
        parse.
    """
    if line:
        line = (
            COMMENT_RE.sub('', line)
            .strip()
        )

        if line:
            try:
                Requirement(line)
            except InvalidRequirement:
                logger.warning('Skipping invalid dependency in %s: %r',
                               path, line)
                line = ''

    return line



"""Patches to setuptools to control behavior.

Version Added:
    1.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from setuptools.config.pyprojecttoml import _ConfigExpander

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence


#: Original initialization function for _ConfigExpander.
_old_ConfigExpander_init = _ConfigExpander.__init__


#: A list of functions to run after _ConfigExpander finishes initializing.
#:
#: Version Added:
#:     1.0
_config_expander_post_funcs: list[Callable[[_ConfigExpander], None]] = []


def patch_setuptools_package_deps(
    deps: Sequence[str],
) -> None:
    """Patch setuptools to set a static list of package dependencies.

    The patch is applied to the next ``_ConfigExpander`` instance that
    setuptools constructs. According to PEP 517, frontends are *supposed* to
    only invoke build hooks once per process. Unfortunately, ``tox`` (via
    ``pyproject_api``) invokes multiple build hooks in succession,
    and each hook causes setuptools to instantiate a fresh
    ``_ConfigExpander``.

    We replace any previously registered patch so the deps from this call apply
    to the next expander only (otherwise stale patches from earlier hooks
    would stack and ``setdefault(...).extend()`` would duplicate the dependency
    list).

    Version Added:
        1.0

    Args:
        deps (list of str):
            The list of dependencies to set.
    """
    # Apply the patch to the next _ConfigExpander instance that setuptools
    # constructs.
    def _patch_deps(
        self: _ConfigExpander,
    ) -> None:
        if 'dependencies' in self.dynamic:
            self.dynamic.remove('dependencies')

        self.dynamic_cfg.pop('dependencies', None)
        self.project_cfg.setdefault('dependencies', []).extend(deps)

    global _config_expander_post_funcs

    _config_expander_post_funcs = [_patch_deps]


def patch_setuptools() -> None:
    """Patch setuptools for buildthings customizations.

    Setuptools isn't built to be customizable, and has some hard-coded
    design elements that it gains from, well, monkey-patching distutils.

    We perform the same kind of patching on Setuptools to control some of
    the :file:`pyproject.toml` loading behavior in order to provide some
    control over the final package builds.

    Version Added:
        1.0
    """
    def _ConfigExpander_init(self, *args, **kwargs) -> None:
        _old_ConfigExpander_init(self, *args, **kwargs)

        for func in _config_expander_post_funcs:
            func(self)

    assert _ConfigExpander.__init__ is _old_ConfigExpander_init, (
        'Attempted to patch _ConfigExpander.__init__ more than once!'
    )

    _ConfigExpander.__init__ = _ConfigExpander_init



"""Tests for buildthings.setuptools_patches.

Version Added:
    1.1
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from setuptools.config.pyprojecttoml import _ConfigExpander

from buildthings import setuptools_patches
from buildthings.setuptools_patches import (
    _old_ConfigExpander_init,
    patch_setuptools,
    patch_setuptools_package_deps,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any


@pytest.fixture(autouse=True)
def patched_setuptools() -> Iterator[None]:
    """Apply the buildthings patch to _ConfigExpander.

    This will apply the patch for the duration of one test, then restore the
    original ``__init__`` and clear any registered post-funcs.

    ``patch_setuptools()`` asserts it hasn't been called before, so we restore
    ``_ConfigExpander.__init__`` to the original between tests.

    Version Added:
        1.1
    """
    saved_init = _ConfigExpander.__init__
    _ConfigExpander.__init__ = _old_ConfigExpander_init
    setuptools_patches._config_expander_post_funcs.clear()

    patch_setuptools()

    try:
        yield
    finally:
        _ConfigExpander.__init__ = saved_init
        setuptools_patches._config_expander_post_funcs.clear()


def _make_pyproject_config(
    deps_dynamic: bool = True,
) -> dict[str, Any]:
    """Return a minimal pyproject.toml-style config for _ConfigExpander.

    The "dependencies" field is marked dynamic by default, mirroring how
    djblets and Review Board's pyproject.toml files declare dependencies that
    buildthings then injects.

    Version Added:
        1.1

    Args:
        deps_dynamic (bool, optional):
            Whether to include dynamic dependencies.

    Returns:
        dict:
        The config to use for the test.
    """
    project: dict[str, Any] = {
        'name': 'example',
    }

    if deps_dynamic:
        project['dynamic'] = ['dependencies']

    return {'project': project}


def test_patch_applies_deps_to_next_expander() -> None:
    """Test that patch is applied to the next _ConfigExpander.

    Version Added:
        1.1
    """
    patch_setuptools_package_deps(['foo>=1', 'bar'])

    expander = _ConfigExpander(_make_pyproject_config())

    assert expander.project_cfg['dependencies'] == ['foo>=1', 'bar']
    assert 'dependencies' not in expander.dynamic


def test_patch_after_prior_expander_does_not_assert() -> None:
    """Test patching after a prior expander has run.

    An expander built before any patch is registered must not poison the
    next ``patch_setuptools_package_deps()`` call.

    :py:mod:`pyproject_api` keeps a single worker process across all PEP 517
    hooks. The earlier ``get_requires_for_build_*`` hooks construct a
    ``Distribution`` (and therefore a ``_ConfigExpander``) before
    ``prepare_metadata_for_build_wheel`` ever calls
    ``patch_setuptools_package_deps()``. The patch call must not raise
    just because some unrelated expander was instantiated earlier.

    Version Added:
        1.1
    """
    # Simulate the earlier hook -- a _ConfigExpander is built before any
    # patch is registered.
    _ConfigExpander(_make_pyproject_config())

    # Now the later hook registers its deps. This must not raise.
    patch_setuptools_package_deps(['foo>=1'])

    # And the next expander -- the one we actually care about -- picks up
    # the deps.
    expander = _ConfigExpander(_make_pyproject_config())

    assert expander.project_cfg['dependencies'] == ['foo>=1']


def test_repeated_patch_calls_do_not_stack() -> None:
    """Test that repeated calls do not stack.

    Calling ``patch_setuptools_package_deps()`` twice replaces the patch rather
    than stacking it.

    Without replacement, both _patch_deps closures would run against the
    next expander, and ``project_cfg.setdefault('dependencies', []).extend()``
    would produce a duplicated dependency list.
    """
    patch_setuptools_package_deps(['foo>=1'])
    patch_setuptools_package_deps(['bar>=2'])

    expander = _ConfigExpander(_make_pyproject_config())

    assert expander.project_cfg['dependencies'] == ['bar>=2']



"""Build backend for buildthings.

Version Added:
    1.0
"""

from __future__ import annotations

import logging
import os
import subprocess
import sys
from typing import TYPE_CHECKING

from setuptools import build_meta as _build_meta
from typing_extensions import TypeAlias

from buildthings.config import IsolationConfig, PyProjectConfig
from buildthings.local_paths import apply_local_dep_paths
from buildthings.requirements import filter_dependencies
from buildthings.setuptools_patches import (patch_setuptools,
                                            patch_setuptools_package_deps)

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from buildthings.config import BuildStep, BuildType

    ConfigSettings: TypeAlias = Mapping[
        str,
        str | list[str] | None,
    ]


# Inject all default hooks from setuptools as defaults into this module.
from setuptools.build_meta import *  # noqa


logger = logging.getLogger(__name__)


#: A cached instance of the loaded pyproject.toml state.
#:
#: Version Added:
#:     1.0
_config: (PyProjectConfig | None) = None


# Patch setuptools so we can inject configuration for the build.
patch_setuptools()


def get_requires_for_build_editable(
    config_settings: (ConfigSettings | None) = None,
) -> Sequence[str]:
    """Return build-time requirements for editable builds.

    This will return the standard dependencies for building the editable
    install by default.

    Additional dependencies can be provided or modified by:

    .. code-block:: toml

       [tool.buildthings.isolation.editable]
       include-dev-deps = true
       include-install-deps = true
       exclude-deps = ['somedep']

    If any of the resulting dependencies are overriden by local dependencies
    found in :file:`.local-packages/`, then the local package paths will
    take precedence.

    Version Added:
        1.0

    Args:
        config_settings (dict, optional):
            Configuration settings to pass to the build backend.

    Returns:
        list of str:
        The list of build-time dependencies.
    """
    return _get_requires_for_build(
        build_type='editable',
        config_settings=config_settings,
    )


def get_requires_for_build_sdist(
    config_settings: (ConfigSettings | None) = None,
) -> Sequence[str]:
    """Return build-time requirements for source distributions.

    This will return the standard dependencies for building the source
    distribution by default.

    Additional dependencies can be provided or modified by:

    .. code-block:: toml

       [tool.buildthings.isolation.sdist]
       include-dev-deps = true
       include-install-deps = true
       exclude-deps = ['somedep']

    If any of the resulting dependencies are overriden by local dependencies
    found in :file:`.local-packages/`, then the local package paths will
    take precedence.

    Version Added:
        1.0

    Args:
        config_settings (dict, optional):
            Configuration settings to pass to Setuptools.

    Returns:
        list of str:
        The list of build-time dependencies.
    """
    return _get_requires_for_build(
        build_type='sdist',
        config_settings=config_settings,
    )


def get_requires_for_build_wheel(
    config_settings: (ConfigSettings | None) = None,
) -> Sequence[str]:
    """Return build-time requirements for wheel distributions.

    This will return the standard dependencies for building the wheel by
    default.

    Additional dependencies can be provided or modified by:

    .. code-block:: toml

       [tool.buildthings.isolation.wheel]
       include-dev-deps = true
       include-install-deps = true
       exclude-deps = ['somedep']

    If any of the resulting dependencies are overriden by local dependencies
    found in :file:`.local-packages/`, then the local package paths will
    take precedence.

    Version Added:
        1.0

    Args:
        config_settings (dict, optional):
            Configuration settings to pass to Setuptools.

    Returns:
        list of str:
        The list of build-time dependencies.
    """
    return _get_requires_for_build(
        build_type='wheel',
        config_settings=config_settings,
    )


def prepare_metadata_for_build_editable(
    metadata_directory: str,
    config_settings: (ConfigSettings | None) = None,
) -> str:
    """Prepare metadata for an editable build.

    This will write out the package's editable dependencies to a temporary
    file so that :file:`pyproject.toml` can reference it.

    Version Added:
        1.0

    Args:
        metadata_directory (str):
            The target directory for metadata.

        config_settings (dict, optional):
            Configuration settings to pass to Setuptools.

    Returns:
        str:
        The basename for the generated ``.dist-info`` directory.
    """
    patch_setuptools_package_deps(_get_editable_package_dependencies())

    return _build_meta.prepare_metadata_for_build_editable(
        metadata_directory,
        config_settings,
    )


def prepare_metadata_for_build_wheel(
    metadata_directory: str,
    config_settings: (ConfigSettings | None) = None,
) -> str:
    """Prepare metadata for a wheel distribution.

    This will write out the package's dependencies to a temporary file
    so that :file:`pyproject.toml` can reference it.

    Version Added:
        1.0

    Args:
        metadata_directory (str):
            The target directory for metadata.

        config_settings (dict, optional):
            Configuration settings to pass to Setuptools.

    Returns:
        str:
        The basename for the generated ``.dist-info`` directory.
    """
    patch_setuptools_package_deps(_get_package_dependencies('wheel'))

    return _build_meta.prepare_metadata_for_build_wheel(
        metadata_directory,
        config_settings,
    )


def build_editable(
    wheel_directory: str,
    config_settings: (ConfigSettings | None) = None,
    metadata_directory: (str | None) = None,
) -> str:
    """Build an editable environment.

    This will perform the following steps:

    1. Write out the editable dependencies to a file for setuptools to load.
    2. If configured, set up the NPM workspaces and install NPM packages.
    3. Run any package-defined build steps for editable mode.
    4. Build the editable install in compat mode (which allows for static
       instead of dynamic lookup of packages, needed for type checkers
       and similar tools).

    Version Added:
        1.0

    Args:
        wheel_directory (str):
            The directory in which to place the editable wheel.

        config_settings (dict, optional):
            Any configuration settings to apply for the build.

        metadata_directory (str, optional):
            A custom path to the ``.dist-info`` directory.

    Returns:
        str:
        The basename of the generated ``.whl`` file.
    """
    config = _get_config()

    return _run_build(
        build_type='editable',
        deps=_get_editable_package_dependencies(),
        dest_directory=wheel_directory,
        config_settings={
            'editable_mode': config.editable_mode,
            **(config_settings or {}),
        },
        metadata_directory=metadata_directory,
    )


def build_sdist(
    sdist_directory: str,
    config_settings: (ConfigSettings | None) = None,
) -> str:
    """Build a source distribution.

    This will perform the following steps:

    1. Write out the package dependencies to a file for setuptools to load.
    2. If configured, set up the NPM workspaces and install NPM packages.
    3. Run any package-defined build steps.
    4. Build the source distribution.

    Version Added:
        1.0

    Args:
        sdist_directory (str):
            The directory where the source distribution will be placed.

        config_settings (dict, optional):
            Any configuration settings to apply for the build.

    Returns:
        str:
        The basename of the generated ``.tar.gz`` file.
    """
    return _run_build(
        build_type='sdist',
        deps=_get_package_dependencies('sdist'),
        dest_directory=sdist_directory,
        config_settings=config_settings,
    )


def build_wheel(
    wheel_directory: str,
    config_settings: (ConfigSettings | None) = None,
    metadata_directory: (str | None) = None,
) -> str:
    """Build a wheel distribution.

    This will perform the following steps:

    1. Write out the package dependencies to a file for setuptools to load.
    2. If configured, set up the NPM workspaces and install NPM packages.
    3. Run any package-defined build steps.
    4. Build the wheel distribution.

    Version Added:
        1.0

    Args:
        wheel_directory (str):
            The directory in which to place the wheel.

        config_settings (dict, optional):
            Any configuration settings to apply for the build.

        metadata_directory (str, optional):
            A custom path to the ``.dist-info`` directory.

    Returns:
        str:
        The basename of the generated ``.whl`` file.
    """
    return _run_build(
        build_type='wheel',
        deps=_get_package_dependencies('wheel'),
        dest_directory=wheel_directory,
        config_settings=config_settings,
        metadata_directory=metadata_directory,
    )


def _get_config() -> PyProjectConfig:
    """Return configuration for the project.

    This will load the configuration on first call and then return that
    instance going forward.

    Version Added:
        1.0

    Returns:
        buildthings.config.PyProjectConfig:
        The :file:`pyproject.toml` configuration file state.
    """
    global _config

    if _config is None:
        _config = PyProjectConfig.from_file()

    return _config


def _get_isolated_build_dependencies(
    isolation_config: IsolationConfig,
    build_type: BuildType,
) -> Sequence[str]:
    """Return the dependencies needed for an isolated build environment.

    These dependencies will be installed in the isolated build environment.

    This doesn't include any dependencies by default, other than those
    specified by ``build-system.requires``, but the following options can
    enable development dependencies and install-time dependencies:

    .. code-block:: toml

       [tool.buildthings.isolation]
       include-dev-deps = true
       include-install-deps = true
       exclude-deps = ['somedep']

    Any packages set to be excluded from installs will then be removed from
    the final list.

    Version Added:
        1.0

    Args:
        isolation_config (buildthings.config.IsolationConfig):
            The configuration for the isolation environment.

        build_type (buildthings.config.BuildType):
            The build type for the environment.

    Returns:
        list of str:
        The list of dependency specifiers.
    """
    config = _get_config()
    deps: list[str] = []

    exclude_deps = set(isolation_config['exclude_deps'])

    if isolation_config['include_dev_deps']:
        deps += config.dev_dependencies

    if isolation_config['include_install_deps']:
        package_deps_config = config.package_dependencies[build_type]

        deps += package_deps_config['dependencies']
        exclude_deps.update(package_deps_config['exclude_deps'])

    if exclude_deps:
        deps = filter_dependencies(deps,
                                   exclude=exclude_deps)

    return deps


def _get_editable_package_dependencies() -> Sequence[str]:
    """Return the dependencies needed for an editable package.

    These dependencies will be installed in the isolated build environment.

    This includes the developer dependencies by default. If the package
    specifies that all dependencies should be included when building the
    package, those will be included as well.

    Any packages set to be excluded from editable installs will then be
    removed from the final list.

    Version Added:
        1.0

    Returns:
        list of str:
        The list of dependency specifiers.
    """
    config = _get_config()
    package_deps_config = config.package_dependencies['editable']

    deps = [
        *config.dev_dependencies,
        *package_deps_config['dependencies'],
    ]

    exclude_deps = package_deps_config['exclude_deps']

    if exclude_deps:
        deps = filter_dependencies(deps,
                                   exclude=set(exclude_deps))

    return deps


def _get_package_dependencies(
    build_type: BuildType,
) -> Sequence[str]:
    """Return the dependencies needed for a package's metadata.

    Any packages set to be excluded will be removed from the final list.

    Version Added:
        1.1

    Args:
        build_type (buildthings.config.BuildType):
            The build type for the environment.

    Returns:
        list of str:
        The list of dependency specifiers.
    """
    config = _get_config()
    package_deps_config = config.package_dependencies[build_type]

    deps = package_deps_config['dependencies']
    exclude_deps = package_deps_config['exclude_deps']

    if exclude_deps:
        deps = filter_dependencies(deps,
                                   exclude=set(exclude_deps))

    return deps


def _get_requires_for_build(
    *,
    build_type: BuildType,
    config_settings: ConfigSettings | None,
) -> Sequence[str]:
    """Return build-time requirements for a build type's isolated environment.

    This will return the standard dependencies for building the build type
    by default.

    Additional dependencies can be provided or modified by:

    .. code-block:: toml

       [tool.buildthings.isolation.editable]
       include-dev-deps = true
       include-install-deps = true
       exclude-deps = ['somedep']

    If any of the resulting dependencies are overriden by local dependencies
    found in :file:`.local-packages/`, then the local package paths will
    take precedence.

    Version Added:
        1.0

    Args:
        build_type (str):
            The build type being built.

        config_settings (dict, optional):
            Configuration settings to pass to the build backend.

    Returns:
        list of str:
        The list of build-time dependencies.
    """
    config = _get_config()
    isolation_config = config.isolation_build_configs[build_type]

    build_meta_func = getattr(_build_meta,
                              f'get_requires_for_build_{build_type}')

    # By default, this will just return what setuptools returns, which is
    # just the dependencies missing when calling setup.py (usually empty),
    # along with any extra dependencies that have been requested for the
    # environment (include-dev-deps and include-install-deps settings).
    return apply_local_dep_paths(
        local_packages_path=isolation_config['local_packages_path'],
        deps=[
            *_get_isolated_build_dependencies(isolation_config, build_type),
            *build_meta_func(config_settings),
        ],
    )


def _run_build(
    *,
    build_type: BuildType,
    deps: Sequence[str],
    dest_directory: str,
    config_settings: ConfigSettings | None,
    **kwargs,
) -> str:
    """Perform a build.

    This will perform the following steps:

    1. Patch setuptools with the dependencies for the package.
    2. If configured, set up the NPM workspaces and install NPM packages.
    3. Run any package-defined build steps.
    4. Perform the build using the given configuration.

    Version Added:
        1.0

    Args:
        build_type (str):
            The type of build to perform.

        deps (list of str):
            The list of package dependencies to write.

        dest_directory (str):
            The directory in which to place the build.

        config_settings (dict):
            Any configuration settings to apply for the build.

        **kwargs (dict):
            Additional keyword arguments to pass to the parent backend's
            build function.

    Returns:
        str:
        The basename of the generated build file.
    """
    config = _get_config()
    npm_config = config.npm_configs[build_type]

    patch_setuptools_package_deps(deps)

    prev_npm_workspaces: Mapping[str, str] = {}

    try:
        if npm_config['install']:
            prev_npm_workspaces = _rebuild_npm_workspaces(
                python_modules=npm_config['python_modules'],
            )
            _install_npm_packages()

        _run_build_steps(config.extra_build_steps[build_type])

        build_func = getattr(_build_meta, f'build_{build_type}')

        return build_func(dest_directory, config_settings, **kwargs)
    finally:
        # Restore any NPM workspace symlinks.
        for symlink_path, dest_path in prev_npm_workspaces.items():
            try:
                os.unlink(symlink_path)
                os.symlink(dest_path, symlink_path)
            except OSError as e:
                logger.warning('Failed to restore symlink path %r back to '
                               '%r: %s',
                               symlink_path, dest_path, e)


def _rebuild_npm_workspaces(
    *,
    python_modules: Sequence[str],
) -> Mapping[str, str]:
    """Rebuild the links under .npm-workspaces for static media building.

    This will look up the module paths for any Python modules in
    ``tool.buildthings.npm.python-dependencies`` and link them so that
    JavaScript and CSS build infrastructure can import files correctly.

    Version Changed:
        1.0.1:
        This now returns a mapping of previous symlinks.

    Version Added:
        1.0

    Args:
        python_modules (list of str):
            A list of Python modules to inspect and link into the workspace.

    Returns:
        dict:
        A mapping of overridden symlinks to previous destinations.
    """
    logger.info('Setting up NPM workspaces...')

    prev_npm_workspaces: dict[str, str] = {}

    # NOTE: Build backends run from within the root of the source tree.
    #       We could leave this as a relative path, but we'll make it
    #       absolute so there are no surprises.
    npm_workspaces_dir = os.path.join(os.getcwd(), '.npm-workspaces')

    if not os.path.exists(npm_workspaces_dir):
        os.mkdir(npm_workspaces_dir, 0o755)

    from importlib import import_module

    for mod_name in python_modules:
        try:
            mod = import_module(mod_name)
        except ImportError:
            raise RuntimeError(
                f'Missing the dependency for {mod_name}, which is needed to '
                f'compile static media.'
            )

        mod_path = mod.__file__
        assert mod_path

        symlink_path = os.path.join(npm_workspaces_dir, mod_name)

        if os.path.exists(symlink_path):
            # Record the existing symlink destination so we can restore it
            # later.
            prev_npm_workspaces[symlink_path] = os.readlink(symlink_path)

        # Unlink this unconditionally, so we don't have to worry about things
        # like an existing dangling symlink that shows as non-existent.
        try:
            os.unlink(symlink_path)
        except FileNotFoundError:
            pass

        os.symlink(os.path.dirname(mod_path), symlink_path)

    return prev_npm_workspaces


def _install_npm_packages() -> None:
    """Install NPM packages.

    Version Added:
        1.0

    Raises:
        RuntimeError:
            There was an error installing npm packages.
    """
    logger.info('Installing NPM modules...')

    try:
        subprocess.run(
            ['npm', 'install'],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f'Failed to install npm packages: {e.output}')


def _run_build_steps(
    steps: Sequence[BuildStep],
) -> None:
    """Run through build steps defined in the package.

    Each build step will be run within the context of the isolated build
    environment.

    This is supported by all build types.

    Version Added:
        1.0

    Args:
        steps (list of buildthings.config.BuildStep):
            Each build step to run.
    """
    for step in steps:
        command = step['command']
        command = command.replace('{python}', sys.executable)

        label = step.get('label') or f'Running {command}'

        logger.info(label)

        try:
            subprocess.check_call(command,
                                  shell=True)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f'Failed to run command {command!r}: {e}')



"""Tests for buildthings.requirements.filter_dependencies.

Version Added:
    1.1
"""

from __future__ import annotations

from buildthings.requirements import filter_dependencies


def test_with_exclude_empty() -> None:
    """Testing filter_dependencies with excluding empty set.

    Version Added:
        1.1
    """
    deps = filter_dependencies(
        [
            'buildthings',
            'kgb~=7.0',
            'ReviewBoard>=8.0',
        ],
        exclude=set(),
    )

    assert deps == [
        'buildthings',
        'kgb~=7.0',
        'ReviewBoard>=8.0',
    ]


def test_with_exclude_package_name() -> None:
    """Testing filter_dependencies with excluding package name.

    Version Added:
        1.1
    """
    deps = filter_dependencies(
        [
            'buildthings',
            'kgb~=7.0',
            'ReviewBoard>=8.0',
        ],
        exclude={
            'KGB',
        },
    )

    assert deps == [
        'buildthings',
        'ReviewBoard>=8.0',
    ]


def test_with_exclude_extras() -> None:
    """Testing filter_dependencies with excluding [extras].

    Version Added:
        1.1
    """
    deps = filter_dependencies(
        [
            'buildthings',
            'kgb~=7.0',
            'ReviewBoard>=8.0',
            'ReviewBoard[p4]',
        ],
        exclude={
            'buildthings[xyz]',
            'reviewboard[p4]',
        },
    )

    assert deps == [
        'buildthings',
        'kgb~=7.0',
        'ReviewBoard>=8.0',
    ]


def test_with_exclude_all() -> None:
    """Testing filter_dependencies with excluding all packages.

    Version Added:
        1.1
    """
    deps = filter_dependencies(
        [
            'buildthings',
            'kgb~=7.0',
            'ReviewBoard>=8.0',
        ],
        exclude={
            'buildthings',
            'kgb',
            'reviewboard',
        },
    )

    assert deps == []



"""Tests for buildthings.backend.

Version Added:
    1.1
"""

from __future__ import annotations

from importlib.metadata import Distribution

from buildthings.tests.utils import (get_dist_from_wheel,
                                     setup_backend)


#
# Tests for get_requires_for_build_editable.
#

def test_get_requires_for_build_editable() -> None:
    """Test get_requires_for_build_editable.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_editable()

    assert deps == []


def test_get_requires_for_build_editable_with_include_install_deps() -> None:
    """Test get_requires_for_build_editable with include-install-deps=true.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.editable]
        include-install-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_editable()

    assert deps == [
        'Djblets~=100.9',
        'ReviewBoard[p4]',
    ]


def test_get_requires_for_build_editable_with_include_dev_deps() -> None:
    """Test get_requires_for_build_editable with include-dev-deps=true

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.editable]
        include-dev-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_editable()

    assert deps == [
        'grumble',
        'kgb~=7.0',
    ]


def test_get_requires_for_build_editable_with_include_all_deps() -> None:
    """Test get_requires_for_build_editable with include-dev-deps=true and
    include-install-deps=true.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.editable]
        include-install-deps = true
        include-dev-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_editable()

    assert deps == [
        'grumble',
        'kgb~=7.0',
        'Djblets~=100.9',
        'ReviewBoard[p4]',
    ]


def test_get_requires_for_build_editable_with_exclude_deps() -> None:
    """Test get_requires_for_build_editable with exclude-deps.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.editable]
        include-install-deps = true
        include-dev-deps = true
        exclude-deps = ['djblets', 'kgb']
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_editable()

    assert deps == [
        'grumble',
        'ReviewBoard[p4]',
    ]


#
# Tests for get_requires_for_build_sdist.
#

def test_get_requires_for_build_sdist() -> None:
    """Test get_requires_for_build_sdist.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_sdist()

    assert deps == []


def test_get_requires_for_build_sdist_with_include_install_deps() -> None:
    """Test get_requires_for_build_sdist with include-install-deps=true.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.sdist]
        include-install-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_sdist()

    assert deps == [
        'Djblets~=100.9',
        'ReviewBoard[p4]',
    ]


def test_get_requires_for_build_sdist_with_include_dev_deps() -> None:
    """Test get_requires_for_build_sdist with include-dev-deps=true.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.sdist]
        include-dev-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_sdist()

    assert deps == [
        'grumble',
        'kgb~=7.0',
    ]


def test_get_requires_for_build_sdist_with_include_all_deps() -> None:
    """Test get_requires_for_build_sdist with include-dev-deps=true and
    include-install-deps=true.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.sdist]
        include-install-deps = true
        include-dev-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_sdist()

    assert deps == [
        'grumble',
        'kgb~=7.0',
        'Djblets~=100.9',
        'ReviewBoard[p4]',
    ]


def test_get_requires_for_build_sdist_with_exclude_deps() -> None:
    """Test get_requires_for_build_sdist with exclude-deps.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.sdist]
        include-install-deps = true
        include-dev-deps = true
        exclude-deps = ['djblets', 'kgb']
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_sdist()

    assert deps == [
        'grumble',
        'ReviewBoard[p4]',
    ]


#
# Tests for get_requires_for_build_wheel.
#

def test_get_requires_for_build_wheel() -> None:
    """Test get_requires_for_build_wheel.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_wheel()

    assert deps == []


def test_get_requires_for_build_wheel_with_include_install_deps() -> None:
    """Test get_requires_for_build_wheel with include-install-deps=true.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.wheel]
        include-install-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_wheel()

    assert deps == [
        'Djblets~=100.9',
        'ReviewBoard[p4]',
    ]


def test_get_requires_for_build_wheel_with_include_dev_deps() -> None:
    """Test get_requires_for_build_wheel with include-dev-deps=true.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.wheel]
        include-dev-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_wheel()

    assert deps == [
        'grumble',
        'kgb~=7.0',
    ]


def test_get_requires_for_build_wheel_with_include_all_deps() -> None:
    """Test get_requires_for_build_wheel with include-dev-deps=true and
    include-install-deps=true.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.wheel]
        include-install-deps = true
        include-dev-deps = true
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_wheel()

    assert deps == [
        'grumble',
        'kgb~=7.0',
        'Djblets~=100.9',
        'ReviewBoard[p4]',
    ]


def test_get_requires_for_build_wheel_with_exclude_deps() -> None:
    """Test get_requires_for_build_wheel with exclude-deps.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.isolation.wheel]
        include-install-deps = true
        include-dev-deps = true
        exclude-deps = ['djblets', 'kgb']
    """

    with setup_backend(pyproject_toml) as backend:
        deps = backend.get_requires_for_build_wheel()

    assert deps == [
        'grumble',
        'ReviewBoard[p4]',
    ]


#
# Tests for prepare_metadata_for_build_editable.
#

def test_prepare_metadata_for_build_editable() -> None:
    """Test prepare_metadata_for_build_editable.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.prepare_metadata_for_build_editable(
            metadata_directory='.',
        )

        assert path == 'buildthings_test-1.0.dist-info'
        assert Distribution.at(path).requires == [
            'grumble',
            'kgb~=7.0',
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]


def test_prepare_metadata_for_build_editable_with_build_type_deps() -> None:
    """Test prepare_metadata_for_build_editable with build type dependencies
    override.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.editable]
        dependencies = [
            'Djblets>=200',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.prepare_metadata_for_build_editable(
            metadata_directory='.',
        )

        assert path == 'buildthings_test-1.0.dist-info'
        assert Distribution.at(path).requires == [
            'grumble',
            'kgb~=7.0',
            'Djblets>=200',
        ]


def test_prepare_metadata_for_build_editable_with_exclude_deps() -> None:
    """Test prepare_metadata_for_build_editable with build type
    exclude-deps.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
            'ReviewBoard>=90',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.editable]
        exclude-deps = [
            'reviewboard',
            'kgb',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.prepare_metadata_for_build_editable(
            metadata_directory='.',
        )

        assert path == 'buildthings_test-1.0.dist-info'
        assert Distribution.at(path).requires == [
            'grumble',
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]


#
# Tests for prepare_metadata_for_build_wheel.
#

def test_prepare_metadata_for_build_wheel() -> None:
    """Test prepare_metadata_for_build_wheel.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.prepare_metadata_for_build_wheel(
            metadata_directory='.',
        )

        assert path == 'buildthings_test-1.0.dist-info'
        assert Distribution.at(path).requires == [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]


def test_prepare_metadata_for_build_wheel_with_build_type_deps() -> None:
    """Test prepare_metadata_for_build_wheel with build type dependencies
    override.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.wheel]
        dependencies = [
            'Djblets>=200',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.prepare_metadata_for_build_wheel(
            metadata_directory='.',
        )

        assert path == 'buildthings_test-1.0.dist-info'
        assert Distribution.at(path).requires == [
            'Djblets>=200',
        ]


def test_prepare_metadata_for_build_wheel_with_exclude_deps() -> None:
    """Test prepare_metadata_for_build_wheel with build type exclude-deps.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
            'ReviewBoard>=90',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.wheel]
        exclude-deps = [
            'reviewboard',
            'kgb',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.prepare_metadata_for_build_wheel(
            metadata_directory='.',
        )

        assert path == 'buildthings_test-1.0.dist-info'
        assert Distribution.at(path).requires == [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]


#
# Tests for build_editable.
#

def test_build_editable() -> None:
    """Test build_editable.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.build_editable(wheel_directory='.')
        assert path == 'buildthings_test-1.0-0.editable-py3-none-any.whl'

        dist = get_dist_from_wheel(path)

        assert dist.requires == [
            'grumble',
            'kgb~=7.0',
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]


def test_build_editable_with_build_type_deps() -> None:
    """Test build_editable with build type dependencies override.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.editable]
        dependencies = [
            'Djblets>=200',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.build_editable(wheel_directory='.')
        assert path == 'buildthings_test-1.0-0.editable-py3-none-any.whl'

        dist = get_dist_from_wheel(path)

        assert dist.requires == [
            'grumble',
            'kgb~=7.0',
            'Djblets>=200',
        ]


def test_build_editable_with_exclude_deps() -> None:
    """Test build_editable with build type exclude-deps.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
            'ReviewBoard>=90',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.editable]
        exclude-deps = [
            'reviewboard',
            'kgb',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.build_editable(wheel_directory='.')
        assert path == 'buildthings_test-1.0-0.editable-py3-none-any.whl'

        dist = get_dist_from_wheel(path)

        assert dist.requires == [
            'grumble',
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]


#
# Tests for build_wheel.
#

def test_build_wheel() -> None:
    """Test build_wheel.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.build_wheel(wheel_directory='.')
        assert path == 'buildthings_test-1.0-py3-none-any.whl'

        dist = get_dist_from_wheel(path)

        assert dist.requires == [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]


def test_build_wheel_with_build_type_deps() -> None:
    """Test build_wheel with build type dependencies override.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.wheel]
        dependencies = [
            'Djblets>=200',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.build_wheel(wheel_directory='.')
        assert path == 'buildthings_test-1.0-py3-none-any.whl'

        dist = get_dist_from_wheel(path)

        assert dist.requires == [
            'Djblets>=200',
        ]


def test_build_wheel_with_exclude_deps() -> None:
    """Test build_wheel with build type exclude-deps.

    Version Added:
        1.1
    """
    pyproject_toml = """
        [tool.buildthings]
        dependencies = [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
            'ReviewBoard>=90',
        ]
        dev-dependencies = [
            'grumble',
            'kgb~=7.0',
        ]

        [tool.buildthings.wheel]
        exclude-deps = [
            'reviewboard',
            'kgb',
        ]
    """

    with setup_backend(pyproject_toml) as backend:
        path = backend.build_wheel(wheel_directory='.')
        assert path == 'buildthings_test-1.0-py3-none-any.whl'

        dist = get_dist_from_wheel(path)

        assert dist.requires == [
            'Djblets~=100.9',
            'ReviewBoard[p4]',
        ]



"""Basic version and package information."""

from __future__ import annotations


# The version of buildthings.
#
# This is in the format of:
#
#   (Major, Minor, Micro, alpha/beta/rc/final, Release Number, Released)
#
VERSION: tuple[int, int, int, str, int, bool] = \
    (1, 1, 0, 'final', 0, True)


def get_version_string() -> str:
    """Return the display version of buildthings.

    Returns:
        str:
        The display version.
    """
    major, minor, micro, tag, release_num, released = VERSION

    version = f'{major}.{minor}'

    if micro:
        version = f'{version}.{micro}'

    if tag != 'final':
        if tag == 'rc':
            version = f'{version} RC{release_num}'
        else:
            version = f'{version} {tag} {release_num}'

    if not is_release():
        version = f'{version} (dev)'

    return version


def get_package_version() -> str:
    """Return the package version of buildthings.

    Returns:
        str:
        The display version.
    """
    major, minor, micro, tag, release_num, released = VERSION

    version = f'{major}.{minor}'

    if micro:
        version = f'{version}.{micro}'

    if tag != 'final':
        if tag == 'alpha':
            tag = 'a'
        elif tag == 'beta':
            tag = 'b'

        version = f'{version}{tag}{release_num}'

    return version


def is_release() -> bool:
    """Return whether the package is released.

    Returns:
        bool:
        ``True`` if the package is a released build, or ``False`` if it is not.
    """
    return VERSION[5]


__version_info__ = VERSION[:-1]
__version__ = get_package_version()



"""Utilities for applying .local-packages paths to dependency lists.

Version Added:
    1.0
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


#: A cache of computed local paths for dependencies.
#:
#: This will only be around for the duration of the build backend hook
#: invocation. Build frontends are expected to invoke each independent hook
#: in its own subprocess. This means this does not need to worry about
#: caching for each build type.
#:
#: Version Added:
#:     1.0
_local_dep_paths: (Mapping[str, str] | None) = None


def get_local_dep_paths(
    local_packages_path: str,
) -> Mapping[str, str]:
    """Return a mapping of dependencies to local paths.

    Args:
        local_packages_path (str):
            The path to the :file:`.local-packages` directory.

    Returns:
        dict:
        A mapping of normalized dependency name to path.
    """
    global _local_dep_paths

    if _local_dep_paths is None:
        local_paths: dict[str, str] = {}

        if local_packages_path and os.path.exists(local_packages_path):
            for name in os.listdir(local_packages_path):
                local_paths[canonicalize_name(name)] = os.path.abspath(
                    os.readlink(os.path.join(local_packages_path, name)))

        _local_dep_paths = local_paths

    return _local_dep_paths


def apply_local_dep_paths(
    *,
    local_packages_path: str | None,
    deps: Sequence[str],
) -> Sequence[str]:
    """Apply local file paths to a list of dependencies.

    This will compare a list of dependencies to any symlinks in
    :file:`.local-packages`, using canonicalized package names.

    If any are found, the dependency will be replaced with::

        <package> @ file://<path>

    Version Added:
        1.0

    Args:
        local_packages_path (str):
            The path to the :file:`.local-packages` directory.

            If ``None``, local dependencies won't be applied.

        deps (list of str):
            The list of dependencies to process.

    Returns:
        list of str:
        The resulting list of dependencies.
    """
    if (deps and
        local_packages_path and
        (local_paths := get_local_dep_paths(local_packages_path))):
        # If any of the dependencies are overridden by a local development
        # package, replace it with that path.
        new_deps: list[str] = []

        for dep in deps:
            req = Requirement(dep)
            dep_name = canonicalize_name(req.name)

            try:
                package_path = local_paths[dep_name]
            except KeyError:
                new_deps.append(dep)
                continue

            new_deps.append(f'{dep_name} @ file://{package_path}')

        deps = new_deps

    return deps



"""Tests for buildthings.requirements.get_package_name_from_specifier.

Version Added:
    1.1
"""

from __future__ import annotations

import logging

from buildthings.requirements import get_package_name_from_specifier


def test_with_package_name() -> None:
    """Test get_package_name_from_specifier with a package name.

    Version Added:
        1.1
    """
    assert get_package_name_from_specifier('ReviewBoard') == 'reviewboard'


def test_with_version_specifier() -> None:
    """Test get_package_name_from_specifier with a version specifier.

    Version Added:
        1.1
    """
    assert get_package_name_from_specifier('ReviewBoard~=8.0') == 'reviewboard'


def test_with_extras() -> None:
    """Test get_package_name_from_specifier with [extras].

    Version Added:
        1.1
    """
    assert (get_package_name_from_specifier('ReviewBoard[P4]~=8.0') ==
            'reviewboard[p4]')


def test_with_extras_multiple() -> None:
    """Test get_package_name_from_specifier with [extras,extras].

    Version Added:
        1.1
    """
    assert (get_package_name_from_specifier('ReviewBoard[p4,LDAP]~=8.0') ==
            'reviewboard[ldap,p4]')


def test_with_invalid_specifier(caplog) -> None:
    """Test get_package_name_from_specifier with invalid specifier.

    Version Added:
        1.1
    """
    with caplog.at_level(logging.WARNING):
        assert (get_package_name_from_specifier('ReviewBoard[~=8/0') ==
                'ReviewBoard[~=8/0')

    assert "Invalid dependency name 'ReviewBoard[~=8/0'" in caplog.text
